/*
 * relay6_socket.c
 *
 * Copyright (C) 2009  Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author(s): David Cantrell <dcantrell@redhat.com>
 */

/*
 * Copyright (C) NEC Europe Ltd., 2003
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <net/if.h>

#include <glib.h>

#include "relay6_socket.h"

fd_set readfd;
gint fdmax;
relay_socket_t *relaysock = NULL;

/* BEGIN STATIC FUNCTIONS */

static void _send_relay_forw(gpointer data, gpointer user_data) {
    gchar *si = (gchar *) data;
    relay_forw_data_t *relay_forw = (relay_forw_data_t *) user_data;
    struct sockaddr_in6 sin6;
    guint32 count = 0;
    gchar dest_addr[INET6_ADDRSTRLEN];
    gint recvmsglen;
    gchar *recvp = NULL;
    struct cmsghdr *cmsgp = NULL;
    struct in6_pktinfo *in6_pkt = NULL;
    struct msghdr msg;
    relay_interface_t *iface = NULL;
    struct iovec iov[1];

    *(relay_forw->mesg->hc_pointer) = MAXHOPCOUNT;
    memset(&sin6, '\0', sizeof(struct sockaddr_in6));
    sin6.sin6_family = AF_INET6;

    memset(dest_addr, 0, INET6_ADDRSTRLEN);
    g_stpcpy(dest_addr, DH6ADDR_ALLSERVER);

    /* destination address */
    if (inet_pton(AF_INET6, dest_addr, &sin6.sin6_addr) <= 0) {
        g_error("%s: inet_pton() failure", __func__);
        return;
    }

    recvmsglen = CMSG_SPACE(sizeof(struct in6_pktinfo));
    recvp = (gchar *) g_malloc0(recvmsglen * sizeof(gchar));

    if (recvp == NULL) {
        g_error("%s: memory allocation error", __func__);
        abort();
    }

    cmsgp = (struct cmsghdr *) recvp;
    cmsgp->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
    cmsgp->cmsg_level = IPPROTO_IPV6;
    cmsgp->cmsg_type = IPV6_2292PKTINFO;
    in6_pkt = (struct in6_pktinfo *) CMSG_DATA(cmsgp);
    msg.msg_control = (void *) recvp;
    msg.msg_controllen = recvmsglen;

    /* destination address */
    if (inet_pton(AF_INET6, dest_addr, &sin6.sin6_addr) <= 0) {
        g_error("%s: inet_pton() failure", __func__);
        return;
    }

    in6_pkt->ipi6_ifindex = if_nametoindex(si);
    sin6.sin6_scope_id = in6_pkt->ipi6_ifindex;

    g_debug("%s: outgoing device index: %d", __func__,
    in6_pkt->ipi6_ifindex);
    iface = get_interface(in6_pkt->ipi6_ifindex);

    if (iface == NULL) {
        g_error("%s: no interface found", __func__);
        exit(0);
    }

    if (inet_pton(AF_INET6, (gchar *) iface->ipv6addr->data,
                  &in6_pkt->ipi6_addr) <= 0) {
        /* source address */
        g_error("%s: inet_pton() failure", __func__);
        abort();
    }

    g_debug("%s: source address: %s", __func__,
            (gchar *) iface->ipv6addr->data);

    sin6.sin6_port = htons(DH6PORT_UPSTREAM_PORT);

    iov[0].iov_base = relay_forw->mesg->buffer;
    iov[0].iov_len = relay_forw->mesg->datalength;
    msg.msg_name = (void *) &sin6;
    msg.msg_namelen = sizeof(sin6);
    msg.msg_iov = &iov[0];
    msg.msg_iovlen = 1;

    if ((count = sendmsg(relaysock->sock_desc, &msg, 0)) < 0) {
        g_error("%s: sendmsg failure: %s", __func__, strerror(errno));
        return;
    }

    if (count > MAX_DHCP_MSG_LENGTH) {
        g_error("%s: sendmsg sent %d bytes while MAX_DHCP_MSG_LENGTH is %d",
                __func__, count, MAX_DHCP_MSG_LENGTH);
    }

    g_debug("%s: => relay_forw, sent to: %s snet_bytes: %d",
            __func__, dest_addr, count);

    g_free(recvp);
    recvp = NULL;
    relay_forw->hit = TRUE;
    return;
}

/* END STATIC FUNCTIONS */

void init_socket(void) {
    relaysock = (relay_socket_t *) g_malloc0(sizeof(relay_socket_t));

    if (relaysock == NULL) {
        g_error("%s: memory allocation error", __func__);
        exit(1);
    }

    relaysock->databuf = (gchar *) g_malloc0(MAX_DHCP_MSG_LENGTH*sizeof(gchar));

    if (relaysock->databuf == NULL) {
        g_error("%s: memory allocation error", __func__);
        exit(1);
    }

    if ((relaysock->sock_desc = socket(AF_INET6, SOCK_DGRAM, 0)) < 0) {
        g_error("%s: failed to get new socket", __func__);
        exit(0);
    }
}

gint get_recv_data(void) {
    struct cmsghdr *cm;
    struct in6_pktinfo *pi;
    struct sockaddr_in6 dst;

    memset(relaysock->src_addr, 0, sizeof(relaysock->src_addr));

    for (cm = (struct cmsghdr *) CMSG_FIRSTHDR(&relaysock->msg); cm;
         cm = (struct cmsghdr *) CMSG_NXTHDR(&relaysock->msg, cm)) {
        if ((cm->cmsg_level == IPPROTO_IPV6)
            && (cm->cmsg_type == IPV6_2292PKTINFO)
            && (cm->cmsg_len == CMSG_LEN(sizeof(struct in6_pktinfo)))) {
            pi = (struct in6_pktinfo *) (CMSG_DATA(cm));
            dst.sin6_addr = pi->ipi6_addr;
            relaysock->pkt_interface = pi->ipi6_ifindex;        /* the
                                                                 * interface
                                                                 * index the
                                                                 * packet got
                                                                 * in */

            if (IN6_IS_ADDR_LOOPBACK(&relaysock->from.sin6_addr)) {
                g_error("%s: source address is loopback", __func__);
                return 0;
            }

            if (inet_ntop(AF_INET6, &relaysock->from.sin6_addr,
                          relaysock->src_addr, INET6_ADDRSTRLEN) <= 0) {
                g_error("%s: inet_ntop failure", __func__);
                return 0;
            }

            if (IN6_IS_ADDR_LOOPBACK(&dst.sin6_addr)) {
                relaysock->dst_addr_type = 1;
            } else if (IN6_IS_ADDR_MULTICAST(&dst.sin6_addr)) {
                relaysock->dst_addr_type = 2;

                if (!multicast) {
                    g_error("%s: received multicast packet is dropped, "
                            "only unicast is allowed", __func__);
                    return 0;
                }
            } else if (IN6_IS_ADDR_LINKLOCAL(&dst.sin6_addr)) {
                relaysock->dst_addr_type = 3;
            } else if (IN6_IS_ADDR_SITELOCAL(&dst.sin6_addr)) {
                relaysock->dst_addr_type = 4;
            }

            return 1;
        }
    }                           /* for */

    return 0;
}

gint check_select(void) {
    gint i = 0;
    gint flag = 0;
    struct timeval tv;

    tv.tv_sec = 0;
    tv.tv_usec = 0;

    FD_ZERO(&readfd);
    /* check the max of them if many desc used */
    fdmax = relaysock->sock_desc;
    FD_SET(fdmax, &readfd);

    if ((i = select(fdmax + 1, &readfd, NULL, NULL, &tv)) == -1) {
        g_error("%s: select() failure", __func__);
        return 0;
    }

    if (FD_ISSET(fdmax, &readfd)) {
        flag = 1;
    } else {
        flag = 0;
    }

    return flag;
}

gint set_sock_opt(void) {
    gint on = 1;
    gint hop_limit;
    relay_interface_t *device = NULL;
    gint flag;
    struct ipv6_mreq sock_opt;
    GSList *iterator = relay_interface_list;

    /* If the relay agent relays messages to the All_DHCP_Servers multicast
     * address or other multicast addresses, it sets the Hop Limit field to
     * 32. [RFC3315 Section 20] */
    hop_limit = 32;
    if (setsockopt(relaysock->sock_desc, IPPROTO_IPV6,
                   IPV6_MULTICAST_HOPS, &hop_limit, sizeof(hop_limit)) < 0) {
        g_error("%s: failed to set socket for IPV6_MULTICAST_HOPS", __func__);
        return 0;
    }

    if (setsockopt(relaysock->sock_desc, IPPROTO_IPV6, IPV6_2292PKTINFO,
                   &on, sizeof(on)) < 0) {
        g_error("%s: failed to set socket for IPV6_2292PKTINFO", __func__);
        return 0;
    }

    while (iterator) {
        device = (relay_interface_t *) iterator->data;
        flag = 0;

        if (g_slist_find_custom(cifaces_list, device->ifname,
                                _find_string) != NULL) {
            flag = 1;
            break;
        }

        if (flag == 0) {
            continue;
        }

        sock_opt.ipv6mr_interface = device->devindex;

        if (inet_pton(AF_INET6, DH6ADDR_ALLAGENT,
                      &sock_opt.ipv6mr_multiaddr) <= 0) {
            g_error("%s: failed to set struct for multicast receive", __func__);
            return 0;
        }

        if (setsockopt(relaysock->sock_desc, IPPROTO_IPV6, IPV6_JOIN_GROUP,
                       (gchar *) &sock_opt, sizeof(sock_opt)) < 0) {
            g_error("%s: failed to set socket option for IPV6_JOIN_GROUP",
                    __func__);
            return 0;
        }

        iterator = g_slist_next(iterator);
    }

    g_debug("%s: socket options are set", __func__);
    return 1;
}


gint fill_addr_struct(void) {
    memset((gchar *) &relaysock->from, 0, sizeof(struct sockaddr_in6));

    relaysock->from.sin6_family = AF_INET6;
    relaysock->from.sin6_addr = in6addr_any;
    relaysock->from.sin6_port = htons(DH6PORT_UPSTREAM_PORT);

    relaysock->iov[0].iov_base = relaysock->databuf;
    relaysock->iov[0].iov_len = MAX_DHCP_MSG_LENGTH;
    relaysock->msg.msg_name = (void *) &relaysock->from;
    relaysock->msg.msg_namelen = sizeof(relaysock->from);
    relaysock->msg.msg_iov = &relaysock->iov[0];
    relaysock->msg.msg_iovlen = 1;

    relaysock->recvmsglen = CMSG_SPACE(sizeof(struct in6_pktinfo));
    relaysock->recvp = (gchar *) g_malloc0(relaysock->recvmsglen*sizeof(gchar));
    relaysock->msg.msg_control = (void *) relaysock->recvp;
    relaysock->msg.msg_controllen = relaysock->recvmsglen;

    if (bind(relaysock->sock_desc, (struct sockaddr *) &relaysock->from,
             sizeof(relaysock->from)) < 0) {
        perror("bind");
        return 0;
    }

    return 1;
}

gint recv_data(void) {
    gint count = -1;

    memset(relaysock->databuf, 0, (MAX_DHCP_MSG_LENGTH * sizeof(gchar)));

    if ((count = recvmsg(relaysock->sock_desc, &relaysock->msg, 0)) < 0) {
        g_error("%s: failed to receive data with recvmsg()", __func__);
        return -1;
    }

    relaysock->buflength = count;

    return 1;
}

gint get_interface_info(void) {
    FILE *f = NULL;
    gchar *addr6 = NULL;
    gchar devname[20];
    struct sockaddr_in6 sap;
    gint plen, scope, dad_status, if_idx;
    gchar addr6p[8][5];
    gchar src_addr[INET6_ADDRSTRLEN];
    relay_interface_t *device = NULL;
    gint opaq = OPAQ;
    gint sw = 0;
    GSList *iterator = NULL;

    if ((f = fopen(INTERFACEINFO, "r")) == NULL) {
        g_error("%s: could not open file", __func__);
        return 0;
    }

    while (fscanf(f, "%4s%4s%4s%4s%4s%4s%4s%4s %02x %02x %02x %02x %20s\n",
                  addr6p[0], addr6p[1], addr6p[2], addr6p[3], addr6p[4],
                  addr6p[5], addr6p[6], addr6p[7], &if_idx, &plen, &scope,
                  &dad_status, devname) != EOF) {
        GString *tmp = g_string_new(NULL);
        g_string_printf(tmp, "%s:%s:%s:%s:%s:%s:%s:%s", addr6p[0],
                        addr6p[1], addr6p[2], addr6p[3], addr6p[4],
                        addr6p[5], addr6p[6], addr6p[7]);
        addr6 = g_strdup(tmp->str);

        if (g_string_free(tmp, TRUE) != NULL) {
            g_error("%s: erroring releasing temporary GString", __func__);
        }

        memset(src_addr, 0, INET6_ADDRSTRLEN);
        sap.sin6_family = AF_INET6;
        sap.sin6_port = 0;

        if (inet_pton(AF_INET6, addr6, sap.sin6_addr.s6_addr) <= 0) {
            g_free(addr6);
            return 0;
        }

        if (inet_ntop(AF_INET6, &sap.sin6_addr, src_addr,
                      sizeof(src_addr)) <= 0) {
            g_free(addr6);
            return 0;
        }

        if (IN6_IS_ADDR_LOOPBACK(&sap.sin6_addr)) {
            g_free(addr6);
            continue;
        }

        sw = 0;
        iterator = relay_interface_list;

        while (iterator) {
            device = (relay_interface_t *) iterator->data;

            if (device->devindex == if_idx) {
                sw = 1;
                break;
            }

            iterator = g_slist_next(iterator);
        }

        if (sw == 0) {
            opaq += 10;
            device = (relay_interface_t *) g_malloc0(sizeof(relay_interface_t));

            if (device == NULL) {
                g_error("%s: memory allocation error", __func__);
                exit(1);
            }

            device->opaq = opaq;
            device->ifname = g_strdup(devname);
            device->devindex = if_idx;
            device->ipv6addr = NULL;

            relay_interface_list = g_slist_append(relay_interface_list, device);
        }

        if (IN6_IS_ADDR_LINKLOCAL(&sap.sin6_addr)) {
            device->link_local = g_strdup(src_addr);
            g_debug("%s: devname: %s, index: %d, link local addr: %s",
                    __func__, devname, if_idx, src_addr);
        } else {
            device->ipv6addr = g_slist_append(device->ipv6addr,
                                              g_strdup(src_addr));
        }

        if (addr6) {
            g_free(addr6);
            addr6 = NULL;
        }
    }

    iterator = relay_interface_list;

    while (iterator) {
        GSList *remove = NULL;
        device = (relay_interface_t *) iterator->data;

        if (!g_slist_length(device->ipv6addr)) {
            g_debug("%s: remove interface %s as it does not have any "
                    "global address", __func__, device->ifname);
            remove = iterator;
            relay_interface_list = g_slist_remove_all(relay_interface_list,
                                                      remove);
        }

        iterator = g_slist_next(iterator);

        if (remove != NULL) {
            g_free(remove);
            remove = NULL;
        }
    }

    fclose(f);
    return 1;
}

gint send_message(void) {
    struct sockaddr_in6 sin6;   /* my address information */
    struct msghdr msg;
    guint32 count = 0;
    struct in6_pktinfo *in6_pkt;
    struct cmsghdr *cmsgp;
    gchar dest_addr[INET6_ADDRSTRLEN];
    relay_interface_t *iface = NULL;
    struct iovec iov[1];
    gint recvmsglen;
    gchar *recvp = NULL;
    GSList *uservers = NULL;
    relay_forw_data_t relay_forw;

    relay_forw.hit = FALSE;

    if ((relay_forw.mesg = get_send_messages_out()) == NULL) {
        return 0;
    }

    if (relay_forw.mesg->sent == 1) {
        return 0;
    }

    memset(&sin6, '\0', sizeof(struct sockaddr_in6));
    sin6.sin6_family = AF_INET6;
    sin6.sin6_flowinfo = 0;
    sin6.sin6_scope_id = 0;

    if (relay_forw.mesg->msg_type == DH6_RELAY_REPL) {
        memset(dest_addr, 0, INET6_ADDRSTRLEN);
        memcpy(dest_addr, relay_forw.mesg->peer_addr, INET6_ADDRSTRLEN);

        recvmsglen = CMSG_SPACE(sizeof(struct in6_pktinfo));
        recvp = (gchar *) g_malloc0(recvmsglen * sizeof(gchar));

        if (recvp == NULL) {
            g_error("%s: memory allocation error", __func__);
            exit(1);
        }

        cmsgp = (struct cmsghdr *) recvp;
        cmsgp->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
        cmsgp->cmsg_level = IPPROTO_IPV6;
        cmsgp->cmsg_type = IPV6_2292PKTINFO;
        in6_pkt = (struct in6_pktinfo *) CMSG_DATA(cmsgp);
        msg.msg_control = (void *) recvp;
        msg.msg_controllen = recvmsglen;

        /* destination address */
        if (inet_pton(AF_INET6, dest_addr, &sin6.sin6_addr) <= 0) {
            g_error("%s: inet_pton() failure", __func__);
            exit(1);
        }

        sin6.sin6_scope_id = relay_forw.mesg->if_index;

        if (relay_forw.mesg->hop > 0) {
            sin6.sin6_port = htons(DH6PORT_UPSTREAM_PORT);
        } else {
            sin6.sin6_port = htons(DH6PORT_DOWNSTREAM_PORT);
        }

        iface = get_interface(relay_forw.mesg->if_index);

        if (iface != NULL) {
            gchar *src_addr;

            if (IN6_IS_ADDR_LINKLOCAL(&sin6.sin6_addr)) {
                src_addr = iface->link_local;
            } else {
                src_addr = (gchar *) iface->ipv6addr->data;
            }

            if (inet_pton(AF_INET6, src_addr, &in6_pkt->ipi6_addr) <= 0) {
                /* source address */
                g_error("%s: inet_pton() failure", __func__);
                exit(1);
            }

            g_debug("%s: source address: %s", __func__, src_addr);
        } else {
            /* the kernel will choose the source address */
            memset(&in6_pkt->ipi6_addr, 0, sizeof(in6_pkt->ipi6_addr));
        }

        /* OUTGOING DEVICE FOR RELAY_REPLY MSG */
        in6_pkt->ipi6_ifindex = relay_forw.mesg->if_index;
        g_debug("%s: outgoing device index: %d", __func__,
                in6_pkt->ipi6_ifindex);
        g_debug("%s: destination port: %d", __func__,
              ntohs(sin6.sin6_port));

        iov[0].iov_base = relay_forw.mesg->buffer;
        iov[0].iov_len = relay_forw.mesg->datalength;
        msg.msg_name = (void *) &sin6;
        msg.msg_namelen = sizeof(sin6);
        msg.msg_iov = &iov[0];
        msg.msg_iovlen = 1;

        if ((count = sendmsg(relaysock->sock_desc, &msg, 0)) < 0) {
            perror("sendmsg");
            return 0;
        }

        if (count > MAX_DHCP_MSG_LENGTH) {
            perror("bytes in sendmsg");
        }

        g_debug("%s: *> relay_repl, sent to: %s sent_bytes: %d",
                __func__, dest_addr, count);

        g_free(recvp);
        recvp = NULL;

        relay_forw.mesg->sent = 1;
        return 1;
    }

    if (relay_forw.mesg->msg_type == DH6_RELAY_FORW) {
        GSList *iterator = IPv6_uniaddr_list;

        while (iterator) {
            gchar *ipv6uni = (gchar *) iterator->data;
            memset(&sin6, '\0', sizeof(struct sockaddr_in6));
            sin6.sin6_family = AF_INET6;

            memset(dest_addr, 0, INET6_ADDRSTRLEN);
            memcpy(dest_addr, ipv6uni, INET6_ADDRSTRLEN);

            /* destination address */
            if (inet_pton(AF_INET6, dest_addr, &sin6.sin6_addr) <= 0) {
                g_error("%s: inet_pton() failure", __func__);
                return 0;
            }

            recvmsglen = CMSG_SPACE(sizeof(struct in6_pktinfo));
            recvp = (gchar *) g_malloc0(recvmsglen * sizeof(gchar));

            if (recvp == NULL) {
                g_error("%s: memory allocation error", __func__);
                exit(1);
            }

            cmsgp = (struct cmsghdr *) recvp;
            cmsgp->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
            cmsgp->cmsg_level = IPPROTO_IPV6;
            cmsgp->cmsg_type = IPV6_2292PKTINFO;
            in6_pkt = (struct in6_pktinfo *) CMSG_DATA(cmsgp);
            msg.msg_control = (void *) recvp;
            msg.msg_controllen = recvmsglen;

            /* destination address */
            if (inet_pton(AF_INET6, dest_addr, &sin6.sin6_addr) <= 0) {
                g_error("%s: inet_pton() failure", __func__);
                return 0;
            }

            sin6.sin6_scope_id = 0;
            sin6.sin6_port = htons(DH6PORT_UPSTREAM_PORT);

            /* the kernel will choose the source address */
            memset(&in6_pkt->ipi6_addr, 0, sizeof(in6_pkt->ipi6_addr));
            /* OUTGOING DEVICE FOR RELAY_REPLY MSG */
            in6_pkt->ipi6_ifindex = 0;

            iov[0].iov_base = relay_forw.mesg->buffer;
            iov[0].iov_len = relay_forw.mesg->datalength;
            msg.msg_name = (void *) &sin6;
            msg.msg_namelen = sizeof(sin6);
            msg.msg_iov = &iov[0];
            msg.msg_iovlen = 1;

            if ((count = sendmsg(relaysock->sock_desc, &msg, 0)) < 0) {
                perror("sendmsg");
                return 0;
            }

            if (count > MAX_DHCP_MSG_LENGTH) {
                perror("bytes sendmsg");
            }

            g_debug("%s: => relay_forw, sent to: %s sent_bytes: %d",
                    __func__, dest_addr, count);
            g_free(recvp);
            recvp = NULL;
            relay_forw.hit = TRUE;

            iterator = g_slist_next(iterator);
        }

        iterator = relay_interface_list;

        while (iterator) {
            uservers = (GSList *) iface->sname;

            while (uservers) {
                gchar *serv = (gchar *) uservers->data;

                memset(&sin6, '\0', sizeof(struct sockaddr_in6));
                sin6.sin6_family = AF_INET6;

                memset(dest_addr, 0, INET6_ADDRSTRLEN);
                memcpy(dest_addr, serv, INET6_ADDRSTRLEN);

                /* destination address */
                if (inet_pton(AF_INET6, dest_addr, &sin6.sin6_addr) <= 0) {
                    g_error("%s: inet_pton() failure", __func__);
                    exit(1);
                }

                recvmsglen = CMSG_SPACE(sizeof(struct in6_pktinfo));
                recvp = (gchar *) g_malloc0(recvmsglen * sizeof(gchar));

                if (recvp == NULL) {
                    g_error("%s: memory allocation error", __func__);
                    exit(1);
                }

                cmsgp = (struct cmsghdr *) recvp;
                cmsgp->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
                cmsgp->cmsg_level = IPPROTO_IPV6;
                cmsgp->cmsg_type = IPV6_2292PKTINFO;
                in6_pkt = (struct in6_pktinfo *) CMSG_DATA(cmsgp);
                msg.msg_control = (void *) recvp;
                msg.msg_controllen = recvmsglen;

                /* destination address */
                if (inet_pton(AF_INET6, dest_addr, &sin6.sin6_addr) <= 0) {
                    g_error("%s: inet_pton() failure", __func__);
                    return 0;
                }

                in6_pkt->ipi6_ifindex = iface->devindex;
                sin6.sin6_scope_id = in6_pkt->ipi6_ifindex;

                g_debug("%s: outgoing device index: %d", __func__,
                        in6_pkt->ipi6_ifindex);
                if (inet_pton(AF_INET6, (gchar *) iface->ipv6addr->data,
                              &in6_pkt->ipi6_addr) <= 0) {
                    /* source address */
                    g_error("%s: inet_pton() failure", __func__);
                    exit(1);
                }

                g_debug("%s: source address: %s", __func__,
                        (gchar *) iface->ipv6addr->data);

                sin6.sin6_port = htons(DH6PORT_UPSTREAM_PORT);

                iov[0].iov_base = relay_forw.mesg->buffer;
                iov[0].iov_len = relay_forw.mesg->datalength;
                msg.msg_name = (void *) &sin6;
                msg.msg_namelen = sizeof(sin6);
                msg.msg_iov = &iov[0];
                msg.msg_iovlen = 1;

                if ((count = sendmsg(relaysock->sock_desc, &msg, 0)) < 0) {
                    perror("sendmsg");
                    return 0;
                }

                if (count > MAX_DHCP_MSG_LENGTH) {
                    perror("bytes sendmsg");
                }

                g_debug("%s: => relay_forw, sent to: %s sent_bytes: %d",
                        __func__, dest_addr, count);
                g_free(recvp);
                recvp = NULL;

                uservers = g_slist_next(uservers);
                relay_forw.hit = TRUE;
            }

            iterator = g_slist_next(iterator);
        }

        g_slist_foreach(sifaces_list, _send_relay_forw, (gpointer) &relay_forw);

        if (relay_forw.hit) {
            GSList *iterator = relay_interface_list;

            while (iterator) {
                iface = (relay_interface_t *) iterator->data;

                if (relay_forw.mesg->interface_in == iface->devindex) {
                    continue;
                }

                *(relay_forw.mesg->hc_pointer) = MAXHOPCOUNT;
                memset(&sin6, '\0', sizeof(struct sockaddr_in6));
                sin6.sin6_family = AF_INET6;

                memset(dest_addr, 0, INET6_ADDRSTRLEN);
                g_stpcpy(dest_addr, DH6ADDR_ALLSERVER);

                /* destination address */
                if (inet_pton(AF_INET6, dest_addr, &sin6.sin6_addr) <= 0) {
                    g_error("%s: inet_pton() failure", __func__);
                    return 0;
                }

                recvmsglen = CMSG_SPACE(sizeof(struct in6_pktinfo));
                recvp = (gchar *) g_malloc0(recvmsglen * sizeof(gchar));

                if (recvp == NULL) {
                    g_error("%s: memory allocation error", __func__);
                    exit(1);
                }

                cmsgp = (struct cmsghdr *) recvp;
                cmsgp->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
                cmsgp->cmsg_level = IPPROTO_IPV6;
                cmsgp->cmsg_type = IPV6_2292PKTINFO;
                in6_pkt = (struct in6_pktinfo *) CMSG_DATA(cmsgp);
                msg.msg_control = (void *) recvp;
                msg.msg_controllen = recvmsglen;

                /* destination address */
                if (inet_pton(AF_INET6, dest_addr, &sin6.sin6_addr) <= 0) {
                    g_error("%s: inet_pton() failure", __func__);
                    return 0;
                }

                sin6.sin6_port = htons(DH6PORT_UPSTREAM_PORT);

                in6_pkt->ipi6_ifindex = iface->devindex;
                sin6.sin6_scope_id = in6_pkt->ipi6_ifindex;

                g_debug("%s: outgoing device index: %d", __func__,
                        in6_pkt->ipi6_ifindex);

                if (inet_pton(AF_INET6, (gchar *) iface->ipv6addr->data,
                              &in6_pkt->ipi6_addr) <= 0) {
                    /* source address */
                    g_error("%s: inet_pton() failure", __func__);
                    exit(1);
                }

                g_debug("%s: source address: %s", __func__,
                        (gchar *) iface->ipv6addr->data);

                iov[0].iov_base = relay_forw.mesg->buffer;
                iov[0].iov_len = relay_forw.mesg->datalength;
                msg.msg_name = (void *) &sin6;
                msg.msg_namelen = sizeof(sin6);
                msg.msg_iov = &iov[0];
                msg.msg_iovlen = 1;

                if ((count = sendmsg(relaysock->sock_desc, &msg, 0)) < 0) {
                    perror("sendmsg");
                    return 0;
                }

                if (count > MAX_DHCP_MSG_LENGTH) {
                    perror("sendmsg");
                }

                g_debug("%s: => relay_forw, sent to: %s sent_bytes: %d",
                        __func__, dest_addr, count);
                g_free(recvp);
                recvp = NULL;

                iterator = g_slist_next(iterator);
            }
        }
    }

    relay_forw.mesg->sent = 1;
    return 1;

    g_error("%s: no message type to send", __func__);
    exit(1);
}
